"""Slot definition."""
import wtrwrks.waterworks.waterwork_part as wp
from wtrwrks.waterworks.empty import empty, Empty
import os

class Slot(wp.WaterworkPart):
  """Object that is always part of some tank which stores the input (in the pour or forward direction) of the operation perfomed by the tank and connects to the tube of another tank.

  Attributes
  ----------
  tank : Tank
    The tank that this slot is a part of.
  key : str
    The string to identify the slot within the tank. Must be unique among all other slots of this tank.
  val : some data type or None
    The value last data inputted to the tank (i.e. operation), if applicable.
  tube : Tube or empty
    The tube from the other tank this tube is connected to, if applicable.
  name : str
    The string used to identify the slot within the entire waterwork. Must be unique among all other slots of this waterwork.
  """
  def __init__(self, tank, key, val=None, tube=empty, plug=None, upstream_slot=None):
    """Initialize the slot.
    Attributes
    ----------
    tank : Tank
      The tank that this slot is a part of.
    key : str
      The string to identify the slot within the tank. Must be unique among all other slots of this tank.
    val : some data type or None
      The value last inputted to the tank (i.e. operation), if applicable.
    tube : Tube or None
      The tube from the other tank this tube is connected to, if applicable.
    """
    self.key = key
    self.tank = tank
    self.tube = tube
    self.name = None
    self.val = val
    self.plug = None
    self.upstream_slot = None

    super(Slot, self).__init__(tank.waterwork, self.name)
    if self.name in self.waterwork.slots:
      raise ValueError(self.name + " already defined as slot. Choose a different name.")
    if plug is not None:
      self.set_plug(plug)
  def __hash__(self):
    """Determine whether two slots are the same within one waterwork."""
    return hash((self.tank, self.key))

  def __eq__(self, other):
    """Determine whether two slots are the same within one waterwork."""
    return (self.tank, self.key) == (other.tank, other.key)

  def __str__(self):
    """Get a string of the name of the slot."""
    return self.name
  #   return str((str(self.tank), str(self.key)))

  def _get_default_name(self, prefix=''):
    """Set a default name. Must be defined by subclass."""
    return os.path.join(self.tank.name, 'slots', self.key)

  def _save_dict(self):
    save_dict = {}
    save_dict['key'] = self.key
    save_dict['tube'] = None if type(self.tube) is Empty else self.tube.name
    save_dict['name'] = self.name
    save_dict['tank'] = self.tank.name
    save_dict['plug'] = self.plug

    return save_dict

  def get_tuple(self):
    """Get a tuple that describes the slot."""
    return (self.tank.name, self.key)

  def get_val(self):
    """Get the value stored in the slot."""
    return self.val

  def set_val(self, val):
    """Set the value stored in the slot."""
    self.val = val

  def set_plug(self, plug, obj_is_callable=False):
    """Plug up the funnel with a function."""

    if self.name not in self.waterwork.funnels:
      raise ValueError("Can only plug funnels. " + str(self) + " is not a funnel.")

    if not callable(plug) or obj_is_callable:
      func_plug = lambda a: plug
    else:
      func_plug = plug

    self.plug = func_plug

  def set_name(self, name):
    """Set the name of the slot within the waterwork."""
    old_name = self.name
    self.name = name

    full_name_space = self.name_space._get_name_string()
    if full_name_space:
      full_name_space = full_name_space + '/'

    if type(name) not in (str, unicode):
      raise TypeError("'name' must be of type str or unicode. Got " + str(type(name)))
    elif not self.name.startswith(full_name_space):
      self.name = os.path.join(self.name_space._get_name_string(), self.name)

    if self.name in self.waterwork.slots:
      raise ValueError(self.name + " already defined as slot. Choose a different name.")

    del self.waterwork.slots[old_name]
    self.waterwork.slots[self.name] = self

    if old_name in self.waterwork.funnels:
      del self.waterwork.funnels[old_name]
      self.waterwork.funnels[self.name] = self

  def unplug(self):
    """Remove the plug function from a funnel."""

    self.plug = None



import shutil
import tempfile
import unittest
import wtrwrks.string_manipulations.diff as df
import os
import pandas as pd
import numpy as np

class TestDiff(unittest.TestCase):
  def setUp(self):
    self.temp_dir = tempfile.mkdtemp()

  def tearDown(self):
    shutil.rmtree(self.temp_dir)

  def test_set_transform(self):
    source = "Hey Hey, what's up"
    target = "Hey hey what's happening"
    diff_string = df.get_diff_string(source, target)

    self.assertEqual(diff_string, '[["i", 4, 5, "h"], ["d", 7, 8, ""], ["i", 16, 17, "ha"], ["i", 18, 18, "pening"]]')
    reco = df.reconstruct(source, diff_string)

    self.assertEqual(target, reco)
if __name__ == "__main__":
  unittest.main()



"""Definition of the base Transform class"""
import pandas as pd
import wtrwrks.utils.dir_functions as d
import wtrwrks.utils.array_functions as af
import wtrwrks.waterworks.waterwork as wa
import os
import numpy as np
import tensorflow as tf
import itertools
import wtrwrks.utils.multiprocessing as mh
import wtrwrks.utils.batch_functions as b
import logging
import random
import glob


class Transform(object):
  """Abstract class used to create mappings from raw to vectorized, normalized data and vice versa. These transform store all the information necessary to create a Waterwork object which can do all the necessary reversible alterations to the data and also to transform it back to its original form.

  Parameters
  ----------
  name : str
    The name of the transform.
  dtype : numpy dtype
    The data type the transformed data should have. Defaults to np.float64.
  input_dtype: numpy dtype
    The datatype of the original inputted array.

  Attributes
  ----------
  attribute_dict: dict
    The keys are the attributes of the class while the values are the default values. It's done this way rather than defined in the __init__ because this dictionary also defines what values need to be saved when written to disk, and what values need to be displayed when printed to the terminal.
  required_params: set of strs
    The parameters that must be provided to the transform at definition.
  cols : list of strs
    The column names of the array or dataframe to be transformed.
  num_examples : int
    The number of rows that have passed through the calc_global_values function.
  is_calc_run : bool
    Whether or not calc_global_values has been run for this transform.

  """

  attribute_dict = {'name': '', 'cols': None, 'num_examples': None, 'is_calc_run': False, 'input_dtype': None, 'dtype': None}
  required_params = set(['name'])

  def __init__(self, from_file=None, save_dict=None, **kwargs):
    """Define a transform using a dictionary, file, or by setting the attribute values in kwargs.

    Parameters
    ----------
    from_file : None or str
      The file path of the Tranform that was written to disk.
    save_dict : dict or None
      The dictionary of attributes that completely define a Transform.
    name : str
      The name of the transform.
    dtype : numpy dtype
      The data type the transformed data should have. Defaults to np.float64.
    input_dtype: numpy dtype
      The datatype of the original inputted array.
    **kwargs :
      The keyword arguments that set the values of the attributes defined in the attribute_dict.

    """
    # Build from file
    if from_file is not None:
      save_dict = d.read_from_file(from_file)
      self._from_save_dict(save_dict)
    # Build from dict
    elif save_dict is not None:
      self._from_save_dict(save_dict)
    # Build from scratch
    else:
      # Check for invalid inputs
      attribute_set = set(self.attribute_dict)
      invalid_keys = sorted(set(kwargs.keys()) - attribute_set)
      if invalid_keys:
        raise TypeError("{} got unexpected keyword argument {}".format(self.__class__.__name__, invalid_keys[0]))

      # Set all the attributes passed to the constructor
      for key in self.attribute_dict:
        if key in kwargs:
          setattr(self, key, kwargs[key])
        else:
          setattr(self, key, self.attribute_dict[key])

      # Ensure all required parameters have been set
      for key in sorted(self.required_params):
        if key not in kwargs:
          raise TypeError("Must supply '{}' as an argument".format(key))

    self.waterwork = None

  def __str__(self):
    """Return the stringified values for each of the attributes in attribute list."""
    return str({a: str(getattr(self, a)) for a in self.attribute_dict})

  def __len__(self):
    raise NotImplementedError()

  def _finish_calc(self):
    """Finish up the calc global value process."""
    return

  def _from_save_dict(self, save_dict):
    """Reconstruct the transform object from the dictionary of attributes."""
    for key in self.attribute_dict:
      setattr(self, key, save_dict[key])

  def _get_array_attributes(self, prefix):
    raise NotImplementedError()

  def _get_dataset(self, file_name_pattern, batch_size, num_epochs=None, num_examples=None, filters=None, keep_features=None, drop_features=None, add_tensors=None, num_threads=1, shuffle_buffer_size=10000, random_seed=None):
    """Create the tensoflow dataset object to be used for input into training pipelines.

    Parameters
    ----------
    file_name_pattern : string
      The file name pattern of the tfrecord files. Supports '*', etc.
    batch_size : int
      The number of example to include in a single batch.
    num_epochs : int
      The number times to run through the full dataset of examples. Defaults to running indefinitely.
    num_examples : int
      Number of examples to run in the dataset before terminating. Cannot use when num_epochs is defined
    filters : dict of functions
      Any pre preprocessing filters to put on the dataset. The keys are the filter names, the values are the filter themselves
    keep_features : list of strs
      The features to keep after reading in from the tfrecords. Defaults to all of them.
    drop_features : list of strs
      The features to drop when reading from the tfrecords. Defaults to None. Can only use this or keep_features not both.
    add_tensors : dict of tensors
      Any additional tensors to add into the dataset object.
    num_threads : int
      The number of io threads to use. Defaults to 1, should probably not be more than 3.
    shuffle_buffer_size : int
      How many examples to shuffle together.
    random_seed : int or None
      The seed to set the random number generator

    Returns
    -------
    dataset : tf.dataset
      The dataset object to feed into tensorflow training pipelines.

    """
    # Set the random seed.
    random.seed(random_seed)

    # Get all the tfrecor file names and shuffle them
    file_names = glob.glob(file_name_pattern)
    file_names.sort()
    shuffled_file_names = random.sample(file_names, len(file_names))

    # Define the dataset object from the tfrecord files
    dataset = tf.data.TFRecordDataset(shuffled_file_names)

    # If num steps was given then use that to define how long to run the dataset
    if num_examples is not None:
      dataset = dataset.take(num_examples)
      dataset = dataset.shuffle(
          buffer_size=shuffle_buffer_size
      )
    # Otherwise use the number of epochs
    else:
      # Convert num epochs to integer
      if num_epochs is not None and type(num_epochs) is not int and not (tf.contrib.framework.is_tensor(num_epochs) and num_epochs.dtype is tf.int64):
        logging.warn(
          '%s is not a whole number. Will be converted to %s',
          num_epochs,
          int(num_epochs)
        )
        num_epochs = int(num_epochs)

      # Shuffle the dataset
      s_and_r = tf.data.experimental.shuffle_and_repeat(
          buffer_size=shuffle_buffer_size,
          count=num_epochs
      )
      dataset = dataset.apply(s_and_r)

    # Read and decode the dataset
    dataset = dataset.map(
        lambda se: self.read_and_decode(se, '', keep_features, drop_features),
        num_parallel_calls=num_threads
    )

    # If any filters were put in place filter the values.
    if filters is not None:
      for key in filters:
        dataset = dataset.filter(
          filters[key]
        )

    # Batch out the data
    if batch_size is not None:
      dataset = dataset.batch(batch_size)

    # Add any additional tensors to the dataset.
    if add_tensors is not None:
      for key in add_tensors:
        def _add_tensor(kwargs):
          kwargs[key] = add_tensors[key]
          return kwargs
        dataset = dataset.map(_add_tensor)
    return dataset

  def _maybe_convert_to_df(self, data):
    """Convert to pandas dataframe if data is array, otherwise do nothing."""
    if type(data) is pd.DataFrame:
      return data
    else:
      return pd.DataFrame(
        data=data,
        columns=self.cols,
        index=np.arange(data.shape[0])
      )

  def _nopre(self, to_unprefix, prefix=''):
    """Strip the self.name/prefix from a string or keys of a dictionary.

    Parameters
    ----------
    to_unprefix : str or dict
      The string or dictionary to strip the prefix/self.name  from.
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

    Returns
    -------
    unprefixed : str or dict
      The string with the prefix/self.name stripped or the dictionary with the prefix/self.name stripped from all the keys.

    """
    str_len = len(os.path.join(prefix, self.name) + '/')
    if str_len == 1:
      str_len = 0
    if type(to_unprefix) is not dict:
      return to_unprefix[str_len:]

    r_d = {}
    for key in to_unprefix:
      if type(key) is tuple and type(key[0]) in (str, unicode):
        r_d[(key[0][str_len:], key[1])] = to_unprefix[key]
      elif type(key) in (str, unicode):
        r_d[key[str_len:]] = to_unprefix[key]
      else:
        r_d[key] = to_unprefix[key]
    return r_d

  def _pre(self, to_prefix, prefix=''):
    """Add the name and some additional prefix to the keys in a dictionary or to a string directly.

    Parameters
    ----------
    to_prefix : str or dict
      The string or dictionary to add to the prefix/self.name prefix to.
    prefix : str
      Any additional prefix to give the string/dictionary keys. Defaults to no additional prefix.

    Returns
    -------
    str or dict
      The string with the prefix/self.name added or the dictionary with the prefix/self.name added to all the keys.

    """
    if type(to_prefix) is not dict:
      return os.path.join(prefix, self.name, to_prefix)

    r_d = {}
    for key in to_prefix:
      if type(key) is tuple and type(key[0]) in (str, unicode):
        r_d[(os.path.join(prefix, self.name, key[0]), key[1])] = to_prefix[key]
      elif type(key) in (str, unicode):
        r_d[os.path.join(prefix, self.name, key)] = to_prefix[key]
      else:
        r_d[key] = to_prefix[key]
    return r_d

  def _save_dict(self):
    """Create the dictionary of values needed in order to reconstruct the transform."""
    save_dict = {}
    for key in self.attribute_dict:
      obj = getattr(self, key)
      save_dict[key] = obj
    save_dict['__class__'] = str(self.__class__.__name__)
    save_dict['__module__'] = str(self.__class__.__module__)
    return save_dict

  def _start_calc(self):
    """Start the calc global value process."""
    self.num_examples = 0.

  def calc_global_values(self, data=None, data_iter=None):
    """Calculate any values needed for the transform that require information from the entired dataset.

    Parameters
    ----------
    data : np.array or pd.DataFrame
      The entire dataset in the form of a numpy array or a pandas DataFrame. Should have the same columns as the arrays that will be fed to the pour method.
    data_iter : iterator of np.array or pd.DataFrame
      The entire dataset in the form of an iterator of numpy array or a pandas DataFrame. Needed if the dataset is too large to fit in memory. Should have the same columns as the arrays that will be fed to the pour method. Can only use if 'data' is not being used

    """
    if data is not None and data_iter is None:
      data_iter = [data]
    elif data_iter is not None and data is None:
      pass
    else:
      raise ValueError("Must supply exactly one array or array_iter.")

    self._start_calc()
    has_at_least_one = False
    for data_num, data in enumerate(data_iter):
      if data_num == 0:
        has_at_least_one = True
        if type(data) is pd.DataFrame:
          is_df = True
          self.cols = list(data.columns)
        else:
          is_df = False
          self.cols = [self.name + '_' + str(dim) for dim in xrange(data.shape[1])]
      data = data.values if is_df else data

      if data_num == 0:
        if self.input_dtype is None:
          self.input_dtype = data.dtype

        if len(data.shape) != 2:
          raise ValueError("Only rank 2 arrays are supported for transforms. Got {}".format(len(data.shape)))

      self._calc_global_values(data)
    self._finish_calc()

    if not has_at_least_one:
      raise ValueError("No data was passed to calc_global_values.")
    self.is_calc_run = True

  def define_waterwork(self, array=None, return_tubes=None, prefix=''):
    """Get the waterwork that completely describes the pour and pump transformations.

    Parameters
    ----------
    array : np.ndarray or empty
      The array to be transformed.
    return_tubes : list of str or None
      Tube objects to be returned from the Waterwork object. Only needed if Waterworks are being stiched together.
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

    Returns
    -------
    Waterwork
      The waterwork with all the tanks (operations) added, and names set.

    """
    raise NotImplementedError()

  def examples_to_tap_dict(self, example_dicts, prefix=''):
    """Run the pump transformation on a list of example dictionaries to reconstruct the original array.

    Parameters
    ----------
    example_dicts: list of dicts of arrays or dict of arrays
      The example dictionaries which the arrays associated with a single example.

    Returns
    -------
    tap_dict: dict
      The dictionary all information needed to completely reconstruct the original data.

    """
    arrays_dict = {}

    # If the examples are all broken up into individual rows then put them into
    # one large dictionary of arrays
    if type(example_dicts) is not dict:
      for array_dict in example_dicts:
        for key in array_dict:
          arrays_dict.setdefault(key, [])
          arrays_dict[key].append(array_dict[key])

      for key in arrays_dict:
        arrays_dict[key] = np.stack(arrays_dict[key])
    else:
      arrays_dict.update(example_dicts)

    # Do any necessary reshaping and recasting
    att_dict = self._get_array_attributes(prefix)
    for key in arrays_dict:
      arrays_dict[key] = arrays_dict[key].reshape([-1] + att_dict[key]['shape'])
      arrays_dict[key] = arrays_dict[key].astype(att_dict[key]['np_type'])

    return arrays_dict

  def get_default_array(self, batch_size=1):
    """Get an array of the proper shape and type that matches the input array.

    Parameters
    ----------
    batch_size : int
      The number of example to include in a single batch.

    Returns
    -------
    default_array : np.array
      The numpy array all filled with some default value

    """
    #####################################
    # NOTE: PUT BACK IN CHECK AFTER WORKFLOW IS RERUN
    # assert self.is_calc_run, ("Run calc_global_values before getting default array")
    ####################################
    return af.empty_array([batch_size, len(self.cols)], self.input_dtype)

  def get_dataset_iter(self, file_name_pattern, batch_size, keep_features=None, drop_features=None, add_tensors=None):
    """Create the tensoflow dataset iterator object used iterator over a tensorflow dataset object.

    Parameters
    ----------
    file_name_pattern : string
      The file name pattern of the tfrecord files. Supports '*', etc.
    batch_size : int
      The number of example to include in a single batch.
    add_tensors : dict of tensors
      Any additional tensors to add into the dataset object.

    Returns
    -------
    dataset_iter : tf.data.Iterator
      The dataset object to feed into tensorflow training pipelines.

    """
    dataset = self._get_dataset(file_name_pattern, batch_size, keep_features=keep_features, drop_features=drop_features, add_tensors=add_tensors)

    data_iter = tf.data.Iterator.from_structure(
        dataset.output_types,
        dataset.output_shapes
    )

    return data_iter

  def get_dataset_feed_iter(self, file_name_pattern, batch_size, keep_features=None, drop_features=None, add_tensors=None):
    """Create the tensoflow dataset iterator object used iterator over a tensorflow dataset object. Used in input pipelines where a feed dict is used.

    Parameters
    ----------
    file_name_pattern : string
      The file name pattern of the tfrecord files. Supports '*', etc.
    batch_size : int
      The number of example to include in a single batch.
    add_tensors : dict of tensors
      Any additional tensors to add into the dataset object.

    Returns
    -------
    dataset_iter : tf.data.Iterator
      The dataset object to feed into tensorflow training pipelines.

    """
    dataset = self._get_dataset(file_name_pattern, batch_size, keep_features=keep_features, drop_features=drop_features, add_tensors=add_tensors)
    handle = tf.placeholder(tf.string, shape=[])

    feed_iter = tf.data.Iterator.from_string_handle(
        handle,
        dataset.output_types,
        dataset.output_shapes
    )

    return feed_iter, handle

  def get_dataset_iter_init(self, dataset_iter, file_name_pattern, batch_size, num_epochs=None, num_examples=None, filters=None, keep_features=None, drop_features=None, add_tensors=None, num_threads=1, shuffle_buffer_size=1000, random_seed=None):
    """Create the tensoflow dataset object to be used for input into training pipelines.

    Parameters
    ----------
    dataset_iter : tf.data.Iterator
      The dataset object to feed into tensorflow training pipelines.
    file_name_pattern : string
      The file name pattern of the tfrecord files. Supports '*', etc.
    batch_size : int
      The number of example to include in a single batch.
    num_epochs : int
      The number times to run through the full dataset of examples. Defaults to running indefinitely.
    num_examples : int
      Number of steps to run in the dataset before terminating. Cannot use when num_epochs is defined
    filters : dict of functions
      Any pre preprocessing filters to put on the dataset. The keys are the filter names, the values are the filter themselves
    keep_features : list of strs
      The features to keep after reading in from the tfrecords. Defaults to all of them.
    drop_features : list of strs
      The features to drop when reading from the tfrecords. Defaults to None. Can only use this or keep_features not both.
    add_tensors : dict of tensors
      Any additional tensors to add into the dataset object.
    num_threads : int
      The number of io threads to use. Defaults to 1, should probably not be more than 3.
    shuffle_buffer_size : int
      How many examples to shuffle together.
    random_seed : int or None
      The seed to set the random number generator

    Returns
    -------
    dataset : tf.dataset
      The dataset object to feed into tensorflow training pipelines.

    """
    dataset = self._get_dataset(file_name_pattern, batch_size, num_epochs, num_examples, filters, keep_features, drop_features, add_tensors, num_threads, shuffle_buffer_size, random_seed)

    return dataset_iter.make_initializer(dataset)

  def get_placeholder(self, tap_key, with_batch=True, batch_size=None):
    """Get a placehold associated with some tap of this transform.

    Parameters
    ----------
    tap_key : str
      The name of the tap to get a placeholder of.
    with_batch : bool
      Whether or not to add a batch dimension
    batch_size : int
      The size of a batch. Defaults to None (i.e. variable).

    Returns
    -------
    ph : tf.placeholder
      The created placeholder object.

    """
    att_dict = self._get_array_attributes()

    # Create the proper shape
    if with_batch:
      shape = [batch_size] + att_dict[tap_key]['shape']
    else:
      shape = att_dict[tap_key]['shape']

    # Get the proper dtype
    if att_dict[tap_key]['np_type'] == np.float64:
      dtype = tf.float64
    else:
      dtype = att_dict[tap_key]['tf_type']

    ph = tf.compat.v1.placeholder(
      dtype=dtype,
      shape=shape
    )

    return ph

  def get_schema_dict(self, var_lim=255):
    """Create a dictionary which defines the proper fields which are needed to store the untransformed data in a (postgres) SQL database.

    Parameters
    ----------
    var_lim : int or dict
      The maximum size of strings. If an int is passed then all VARCHAR fields have the same limit. If a dict is passed then each field gets it's own limit. Defaults to 255 for all string fields.

    Returns
    -------
    schema_dict : dict
      Dictionary where the keys are the field names and the values are the SQL data types.

    """
    # Select the SQL datatype by the input_dtype
    if self.input_dtype in (np.int64, np.int32, int):
      db_dtype = 'INTEGER'
    elif self.input_dtype in (np.float64, np.float32, float):
      db_dtype = 'FLOAT'
    elif self.input_dtype in (np.bool, bool):
      db_dtype = 'BOOLEAN'
    elif self.input_dtype in (np.dtype('S'), np.dtype('U'), np.dtype('O')):
      db_dtype = 'VARCHAR'
    elif self.input_dtype in (np.datetime64,):
      db_dtype = 'TIMESTAMP'
    else:
      raise ValueError("{} is not a supported type to be used on the database.".format())

    # Create the schema dictionary.
    schema_dict = {}
    for col in self.cols:
      if db_dtype == 'VARCHAR' and type(var_lim) is dict and col in var_lim:
        schema_dict[col] = db_dtype + '(' + str(var_lim[col]) + ')'
      elif db_dtype == 'VARCHAR' and type(var_lim) is dict:
         schema_dict[col] = db_dtype + '(' + str(255) + ')'
      elif db_dtype == 'VARCHAR':
        schema_dict[col] = db_dtype + '(' + str(var_lim) + ')'
      else:
        schema_dict[col] = db_dtype
    return schema_dict

  def get_waterwork(self, recreate=False):
    """Create the Transform's waterwork or return the one that was already created.

    Parameters
    ----------
    recreate : bool
      Whether or not to force the transform to create a new waterwork.

    Returns
    -------
    Waterwork
      The waterwork object that this transform creates.

    """
    assert self.is_calc_run, ("Run calc_global_values before running the transform")

    if self.waterwork is not None and not recreate:
      return self.waterwork

    with wa.Waterwork(name=self.name) as ww:
      self.define_waterwork()

    self.waterwork = ww
    return ww

  def pour(self, data=None, data_iter=None):
    """Execute the transformation in the pour (forward) direction.

    Parameters
    ----------
    data : np.ndarray
      The numpy array to transform.
    data_iter : iterator of np.array or pd.DataFrame
      The entire dataset in the form of an iterator of numpy array or a pandas DataFrame. Needed if the dataset is too large to fit in memory. Should have the same columns as the arrays that will be fed to the pour method. Can only use if 'data' is not being used
    Returns
    -------
    tap_dict : dict (or iterator of dicts)
      The dictionary of transformed outputs as well as any additional information needed to completely reconstruct the original data. Returns an iterator of dicts if something is passed to 'data_iter' rather than the 'data' argument.

    """
    ww = self.get_waterwork()

    def normalize_and_pour(data):
      if type(data) is pd.DataFrame:
        data = data.values
      funnel_dict = self._pre({'array': data})
      tap_dict = ww.pour(funnel_dict, key_type='str')
      return tap_dict

    if data is not None and data_iter is None:
      return normalize_and_pour(data)
    elif data_iter is not None and data is None:
      return itertools.imap(normalize_and_pour, data_iter)

    else:
      raise ValueError("Must supply exactly one data or data_iter.")

  def pump(self, tap_dict, df=False, index=None):
    """Execute the transformation in the pump (backward) direction.

    Parameters
    ----------
    tap_dict: dict
      The dictionary all information needed to completely reconstruct the original data.

    Returns
    -------
    data : np.ndarray or df
      The original numpy array or pandas dataframe that was poured.

    """
    ww = self.get_waterwork()
    funnel_dict = ww.pump(tap_dict, key_type='str')
    array = funnel_dict[self._pre('array')].astype(self.input_dtype)
    if df:
      if index is None:
        index = np.arange(array.shape[0])
      data = pd.DataFrame(
        data=array,
        index=index,
        columns=self.cols
      )
    else:
      data = array

    return data

  def read_and_decode(self, serialized_example, prefix='', keep_features=None, drop_features=None):
    """Convert a serialized example created from an example dictionary from this transform into a dictionary of shaped tensors for a tensorflow pipeline.

    Parameters
    ----------
    serialized_example : tfrecord serialized example
      The serialized example to read and convert to a dictionary of tensors.
    prefix : str
      A string to prefix the dictionary keys with.
    keep_features : list of strs
      The features to keep after reading in from the tfrecords. Defaults to all of them.
    drop_features : list of strs
      The features to drop when reading from the tfrecords. Defaults to None. Can only use this or keep_features not both.

    Returns
    -------
    features : dict of tensors
      The tensors created by decoding the serialized example

    """
    feature_dict = {}
    shape_dict = {}

    att_dict = self._get_array_attributes(prefix)

    # Define the set of features that will be kept
    if drop_features is None:
      drop_features = []
    if keep_features is not None:
      att_dict = {k: att_dict[k] for k in keep_features}

    for key in att_dict:
      if key in drop_features:
        continue

      # Define the tf feature object and pull out the correct shape
      feature_dict[key] = tf.FixedLenFeature(
        att_dict[key]['size'],
        att_dict[key]['tf_type']
      )
      shape_dict[key] = att_dict[key]['shape']

    # Parse the example
    features = tf.parse_single_example(
      serialized_example,
      features=feature_dict
    )

    # Do any reshaping and cast to float if the numpy dtype is np.float64
    for key in shape_dict:
      features[key] = tf.reshape(features[key], shape_dict[key])
      np_dtype = att_dict[key]['np_type']
      if np_dtype == np.float64:
        features[key] = tf.cast(features[key], dtype=tf.float64)

    return features

  def save_to_file(self, path):
    """Save the transform object to disk."""
    save_dict = self._save_dict()
    d.save_to_file(save_dict, path)

  def tap_dict_to_examples(self, tap_dict, prefix=''):
    """Run the pour transformation on an array to transform it into a form best for ML pipelines. This list of example dictionaries can be easily converted into tf records, but also have all the information needed in order to reconstruct the original array.

    Parameters
    ----------
    tap_dict : dict (or iterator of dicts)
      The dictionary of transformed outputs as well as any additional information needed to completely reconstruct the original data. Returns an iterator of dicts if something is passed to 'data_iter' rather than the 'data' argument.

    Returns
    -------
    list of dicts of features
      The example dictionaries which contain tf.train.Features.

    """
    num_examples = tap_dict[tap_dict.keys()[0]].shape[0]

    # Get the dictionary of attributes (shape, dtype, etc.) of the arrays in
    # pour_outputs.
    att_dict = self._get_array_attributes(prefix)

    # Go through each row and each key of pour_outputs. Flatten the array and
    # convert it into it's proper feature. Return as list of dicts.
    example_dicts = []

    for row_num in xrange(num_examples):
      example_dict = {}

      for key in tap_dict:
        dtype = att_dict[key]['np_type']
        flat = tap_dict[key][row_num].flatten().astype(dtype)
        example_dict[key] = att_dict[key]['feature_func'](flat)

      example_dicts.append(example_dict)

    return example_dicts

  def write_examples(self, data=None, data_iter=None, file_name=None, file_num_offset=0, batch_size=1, num_threads=1, skip_fails=False, skip_keys=None, use_threading=False, serialize_func=None, prefix=''):
    """Pours the arrays then writes the examples to tfrecords in a multithreading manner. It creates one example per 'row', i.e. axis=0 of the arrays. All arrays must have the same axis=0 dimension and must be of a type that can be written to a tfrecord

    Parameters
    ----------
    data : np.array or pd.DataFrame
      The entire dataset in the form of a numpy array or a pandas DataFrame. Should have the same columns as the arrays that will be fed to the pour method.
    data_iter : iterator of np.array or pd.DataFrame
      The entire dataset in the form of an iterator of numpy array or a pandas DataFrame. Needed if the dataset is too large to fit in memory. Should have the same columns as the arrays that will be fed to the pour method. Can only use if 'data' is not being used
    file_name : str
      The name of the tfrecord file to write to. An extra '_<num>' will be added to the name.
    file_num_offset : int
      A number that controls what number will be appended to the file name (so that files aren't overwritten.)
    batch_size : int
      The number of example to include in a single batch.
    num_threads : int
      The number of io threads to use. Defaults to 1, should probably not be more than 3.
    skip_fails : bool
      Whether or not to skip any write failures without error. Defaults to false.
    skip_keys : list of strs
      Any taps that should not be written to examples.
    use_threading : bool
      Whether or not to use multithreading rather than multiprocessing. Defaults to False
    serialize_func : np.array -> list of serialized examples
      The function that handles the transformation from numpy array to serialized example, if the user does not want to use the default one.
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.
    """
    # Make sure only data or data_iter is passed
    if data is not None and data_iter is None:
      data_iter = [data]
    elif data_iter is not None and data is None:
      pass
    else:
      raise ValueError("Must supply exactly one data or data_iter.")

    # If no serialize_func is passed then define the default one.
    if serialize_func is None:
      save_dict = self._save_dict()
      cls = self.__class__

      def serialize_func(data):
        trans = cls(save_dict=save_dict)

        tap_dict = trans.pour(data)
        example_dicts = trans.tap_dict_to_examples(tap_dict, prefix)

        serials = []
        for example_dict in example_dicts:
          example = tf.train.Example(
            features=tf.train.Features(feature=example_dict)
          )
          serials.append(example.SerializeToString())

        return serials

    # If the data iterator is a list, then convert to tuple
    if type(data_iter) in (list, tuple):
      data_iter = (i for i in data_iter)

    # Make sure the file name ends in tfrecord
    if not file_name.endswith('.tfrecord'):
      raise ValueError("file_name must end in '.tfrecord'")

    # Create the directory if it doesn't exist
    dir = '/'.join(file_name.split('/')[:-1])
    if dir:
      d.maybe_create_dir(dir)

    # Batch out the data iterator into batches of batch size
    file_names = []
    for batch_num, batch in enumerate(b.batcher(data_iter, batch_size)):
      logging.info("Serializing batch %s", batch_num)

      # If skip fails then catch any exceptions otherwise just run the batch
      # through the serialize function
      if skip_fails:
        try:
          all_serials = mh.multi_map(serialize_func, batch, num_threads, use_threading)
        except Exception:
          logging.warn("Batched %s failed. Skipping.", batch_num)
          continue
      else:
        all_serials = mh.multi_map(serialize_func, batch, num_threads, use_threading)

      logging.info("Finished serializing batch %s", batch_num)

      # Add to the full list of created files.
      file_num = file_num_offset + batch_num
      fn = file_name.replace('.tfrecord', '_' + str(file_num) + '.tfrecord')
      file_names.append(fn)

      # Write the examples to disk
      logging.info("Writing batch %s", batch_num)
      writer = tf.io.TFRecordWriter(fn)
      for serials in all_serials:
        for serial in serials:
          writer.write(serial)

      # Close the writer
      logging.info("Finished writing batch %s", batch_num)
      writer.close()

    return file_names



import pathos.multiprocessing as mp


def multi_map(func, iterable, num_threads=1, use_threading=False):
  out_list = []
  if num_threads != 1:
    if not use_threading:
      pool = mp.ProcessingPool(num_threads)
    else:
      pool = mp.ThreadPool(num_threads)

  # If multithreading run pool.map, otherwise just run recon_and_pour
  if num_threads != 1:
    out_list = pool.map(func, iterable)

    # processingpool carries over information. Need to terminate a restart
    # to prevent memory leaks.
    pool.terminate()
    if not use_threading:
      pool.restart()
  else:
    for element in iterable:
      out_list.append(func(element))
  return out_list



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestTranspose(th.TestTank):

  def test_two_d(self):
    a = np.array([[0, 1], [2, 3], [4, 5]])
    self.pour_pump(
      td.transpose,
      {
        'a': a,
        'axes': [1, 0]
      },
      {
        'target': np.array([[0, 2, 4], [1, 3, 5]]),
        'axes': [1, 0]
      },
      type_dict={'a': np.ndarray, 'axes': tuple}
    )

  def test_three_d(self):
    a = np.arange(24).reshape((4, 3, 2, 1))
    self.pour_pump(
      td.transpose,
      {
        'a': a,
        'axes': (0, 1, 3, 2)
      },
      {
        'target': np.transpose(a, (0, 1, 3, 2)),
        'axes': (0, 1, 3, 2)
      },
      type_dict={'a': np.ndarray, 'axes': tuple}
    )
    a = np.arange(24).reshape((4, 3, 2, 1))
    self.pour_pump(
      td.transpose,
      {
        'a': a,
        'axes': (3, 0, 1, 2)
      },
      {
        'target': np.transpose(a, (3, 0, 1, 2)),
        'axes': (3, 0, 1, 2)
      },
      type_dict={'a': np.ndarray, 'axes': tuple}
    )

    self.pour_pump(
      td.transpose,
      {
        'a': a,
        'axes': (0, 3, 1, 2)
      },
      {
        'target': np.transpose(a, (0, 3, 1, 2)),
        'axes': (0, 3, 1, 2)
      },
      type_dict={'a': np.ndarray, 'axes': tuple}
    )


if __name__ == "__main__":
    unittest.main()



"""Functions which deal with differences between strings."""
import difflib
import json


def get_diff_string(source_string, target_string):
  """Generate a string which can be used to transform one string into another.

  Parameters
  ----------
  source_string : str
    The string that uses the diff_string to get transformed into the target string.
  target_string : type
    The string that source string is being transformed into.

  Returns
  -------
  str
    A string that describes a sequence of operations on source_string to produce target_string.

  """
  # Split up the strings into lists.
  source_string = list(source_string)
  target_string = list(target_string)

  # Use a sequence matcher to identify all the common substrings of
  # source_string an target_string.
  matcher = difflib.SequenceMatcher(None, source_string, target_string)

  tag_map = {
    'delete': 'd',
    'insert': 'i',
    'replace': 'i'
  }

  # Get out the operation codes which list out all the operations that need to
  # occur in order to transform source_string into target_string.
  diff_string = []
  for tag, i1, i2, j1, j2 in matcher.get_opcodes():
    if tag == 'equal':
      continue

    substring = ''
    if tag in ('insert', 'replace'):
      substring = ''.join(target_string[j1:j2])

    diff_string.append((tag_map[tag], i1, i2, substring))

  return json.dumps(diff_string)


def reconstruct(source_string, diff_string):
  """Generate a string which can be used to transform one string into another.

  Parameters
  ----------
  source_string : str
    The string that uses the diff_string to get transformed into the target string.
  target_string : str
    A string that describes a sequence of operations on source_string to produce target_string.

  Returns
  -------
  str
    The string that source string is transformed into using the diff string.

  """
  target_string = list(source_string)
  for tag, i1, i2, substring in reversed(json.loads(diff_string)):
    if tag == 'd':
      del target_string[i1:i2]
    elif tag == 'i':
      target_string[i1:i2] = substring

  return ''.join(target_string)



"""CatToIndex tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import wtrwrks.utils.array_functions as af
import numpy as np


class CatToIndex(ta.Tank):
  """The CatToIndex class where the cats input is an numpy array. Handles any rank for 'cats'

  Attributes
  ----------
  slot_keys : list of str
    The tank's (operation's) argument keys. They define the names of the inputs to the tank.
  tube_keys : dict(
    keys - strs. The tank's (operation's) output keys. THey define the names of the outputs of the tank
    values - types. The types of the arguments outputs.
  )
    The tank's (operation's) output keys and their corresponding types.

  """
  func_name = 'cat_to_index'
  slot_keys = ['cats', 'cat_to_index_map']
  tube_keys = ['target', 'cat_to_index_map', 'missing_vals', 'input_dtype']
  pass_through_keys = ['cat_to_index_map']

  def _pour(self, cats, cat_to_index_map):
    """Execute the mapping in the pour (forward) direction .

    Parameters
    ----------
    cats : np.ndarray
      The categorical values to be mapped to indices
    cat_to_index_map : dict
      The map from categorical values to indices.

    Returns
    -------
    dict(
      'target': int, float, other non array type
        The result of the sum of 'a' and 'b'.
      'missing_vals': list
        The list of all the cats that were not found in cat_to_index_map.
      'cat_to_index_map' : dict
        The map from categorical values to indices.
    )

    """
    cats = np.array(cats, copy=True)
    shape = cats.shape
    is_float = cats.dtype in (np.float64, np.float32)
    nan_val = -1
    if is_float:
      for key, value in cat_to_index_map.iteritems():
        if np.isnan(key):
          nan_val = value
          break
    # Pull out all the cats which are not in the cat_to_index_map.
    # If you have a nan not in a float, you're going to get
    # unexpected results, so here we force nans not to show up in
    # missing vals if they are in the cat_to_index_map.
    missing_vals = af.empty_array_like(cats)
    if is_float:
      isnan = np.array([True if str(v) == 'nan' else False for v in cats.flatten()]).reshape(cats.shape)
      mask = (~np.isin(cats, cat_to_index_map.keys())) & (~isnan)
      missing_vals[mask] = cats[mask]
    else:
      mask = ~np.isin(cats, cat_to_index_map.keys())
      missing_vals[mask] = cats[mask]

    # Map all the categorical values to indices, setting an index of -1
    # every time an unsupported category is encoutered.
    def safe_map(cat):
      if cat in cat_to_index_map:
        return cat_to_index_map[cat]
      elif is_float and np.isnan(cat):
        return nan_val
      else:
        return -1
    target = np.vectorize(safe_map)(cats)
    target = target.reshape(shape)

    return {'target': target, 'missing_vals': missing_vals, 'cat_to_index_map': cat_to_index_map, 'input_dtype': cats.dtype}

  def _pump(self, target, missing_vals, cat_to_index_map, input_dtype):
    """Execute the mapping in the pump (backward) direction .

    Parameters
    ----------
    target: np.ndarray
      The result of the sum of 'a' and 'b'.
    missing_vals: list
      The list of all the cats that were not found in cat_to_index_map.
    cat_to_index_map : dict
      The map from categorical values to indices.


    Returns
    -------
    dict(
      'cats' : np.ndarray
        The categorical values to be mapped to indices
      'cat_to_index_map' : dict
        The map from categorical values to indices.
    )

    """
    missing_vals = np.array(missing_vals)
    # Convert the cat_to_index_map into an index_to_cat_map, while making
    # sure it is one-to-one. Otherwise it isn't reversible.
    index_to_cat_map = {}
    for k, v in cat_to_index_map.iteritems():
      if v in index_to_cat_map:
        raise ValueError("cat_to_index_map must be one-to-one. " + str(v) + " appears twice.")
      index_to_cat_map[v] = k

    # Need to set the otypes variable for np.vectorize, otherwise it runs
    # it once to figure out the output type of map_back. This screws up
    # missing_vals.pop
    # otype = str if not len(cat_to_index_map) else type(cat_to_index_map.keys()[0])
    cats = np.array([index_to_cat_map[i] if i != -1 else index_to_cat_map[0] for i in target.flatten()])
    cats = cats.reshape(target.shape)

    mask = target == -1
    cats[mask] = missing_vals[mask]

    cats = cats.astype(input_dtype)

    return {'cats': cats, 'cat_to_index_map': cat_to_index_map}


class MultiCatToIndex(CatToIndex):
  """The CatToIndex class where the cats input is an numpy array. Handles any rank for 'cats'

  Attributes
  ----------
  slot_keys : list of str
    The tank's (operation's) argument keys. They define the names of the inputs to the tank.
  tube_keys : dict(
    keys - strs. The tank's (operation's) output keys. THey define the names of the outputs of the tank
    values - types. The types of the arguments outputs.
  )
    The tank's (operation's) output keys and their corresponding types.

  """
  func_name = 'multi_cat_to_index'
  slot_keys = ['cats', 'cat_to_index_maps', 'selector']
  tube_keys = ['target', 'cat_to_index_maps', 'missing_vals', 'input_dtype', 'selector']
  pass_through_keys = ['cat_to_index_maps', 'selector']

  def _pour(self, cats, selector, cat_to_index_maps):
    """Execute the mapping in the pour (forward) direction .

    Parameters
    ----------
    cats : np.ndarray
      The categorical values to be mapped to indices
    cat_to_index_map : dict
      The map from categorical values to indices.

    Returns
    -------
    dict(
      'target': int, float, other non array type
        The result of the sum of 'a' and 'b'.
      'missing_vals': list
        The list of all the cats that were not found in cat_to_index_map.
      'cat_to_index_map' : dict
        The map from categorical values to indices.
    )

    """
    target = -1 * np.ones(cats.shape, dtype=int)
    missing_vals = af.empty_array_like(cats)

    uniques = np.unique(selector)
    for unique in uniques:
      cat_to_index_map = cat_to_index_maps[unique]
      mask = selector == unique

      selected_cats = cats[mask]
      pour_dict = super(MultiCatToIndex, self)._pour(selected_cats, cat_to_index_map)

      target[mask] = pour_dict['target']
      missing_vals[mask] = pour_dict['missing_vals']
    return {'target': target, 'missing_vals': missing_vals, 'cat_to_index_maps': cat_to_index_map, 'selector': selector, 'input_dtype': cats.dtype}

  def _pump(self, target, selector, missing_vals, cat_to_index_maps, input_dtype):
    """Execute the mapping in the pump (backward) direction .

    Parameters
    ----------
    target: np.ndarray
      The result of the sum of 'a' and 'b'.
    missing_vals: list
      The list of all the cats that were not found in cat_to_index_map.
    cat_to_index_map : dict
      The map from categorical values to indices.


    Returns
    -------
    dict(
      'cats' : np.ndarray
        The categorical values to be mapped to indices
      'cat_to_index_map' : dict
        The map from categorical values to indices.
    )

    """
    cats = af.empty_array_like(missing_vals)

    uniques = np.unique(selector)
    masks = []
    cats_list = []
    itemsize = None

    for unique in uniques:
      cat_to_index_map = cat_to_index_maps[unique]
      mask = selector == unique

      selected_target = target[mask]
      selected_missing_vals = missing_vals[mask]

      pump_dict = super(MultiCatToIndex, self)._pump(selected_target, selected_missing_vals,  cat_to_index_map, input_dtype)

      masks.append(mask)
      cats_list.append(pump_dict['cats'])

    if cats.dtype.type in (np.string_, np.unicode_):
      itemsize = np.max([c.dtype.itemsize for c in cats_list] + [1])
      dtype = str(cats.dtype)[:2]
      cats = cats.astype(dtype + str(itemsize))

    for c, mask in zip(cats_list, masks):
      cats[mask] = c

    return {'cats': cats, 'cat_to_index_maps': cat_to_index_maps, 'selector': selector}



"""Simple functions that help out in tank definitions."""
import copy
import numpy as np


def maybe_copy(a):
  """Copy the object 'a' if it is mutable or possible to copy. Otherwise just return the original object.

  Parameters
  ----------
  a : object
      The object to copy

  Returns
  -------
  object
      The (possibly) copied object

  """
  if type(a) in (str, unicode, int, float):
    return a
  if type(a) is np.ndarray:
    return np.array(a, copy=True)
  if type(a) in (list, dict):
    return copy.copy(a)

  return a




def batcher(iterable, batch_size=None):
  if type(iterable) in (list, tuple):
    iterable = (i for i in iterable)

  out_batch = []
  break_out = False

  while True:
    try:
      out_batch.append(next(iterable))
    except StopIteration:
      break_out = True

    if len(out_batch) == batch_size or (break_out and out_batch):
      yield out_batch
      out_batch = []

    if break_out:
      break



import numpy as np
import wtrwrks.tanks.utils as ut
import wtrwrks.waterworks.tank as ta


def create_one_arg_reduce_tank(np_func, class_name, func_name):
  """Create a function which generates the tank instance corresponding to some single argument, reduceean valued numpy function. (e.g. np.isnan). The operation will be reversible but in the most trivial and wasteful manner possible. It will just copy over the original array.

  Parameters
  ----------
  np_func : numpy function
      A numpy function which operates on an array to give another array.
  class_name : str
      The name you'd like to have the Tank class called.

  Returns
  -------
  func
      A function which outputs a tank instance which behaves like the np_func but is also reversible

  """
  temp = func_name

  # Define the tank subclass.
  class TankClass(ta.Tank):
    func_name = temp
    slot_keys = ['a', 'axis']
    tube_keys = ['target', 'axis', 'a']
    pass_through_keys = ['a', 'axis']

    def _pour(self, a, axis):
      # If an empty tuple was given then set the axis to None
      if not np.array(axis).size:
        input_axis = None
      else:
        input_axis = axis
      axis = np.array(axis)

      # Reduce the array using the supplied numpy array function.
      target = np_func(a, axis=input_axis)
      return {'target': target, 'a': ut.maybe_copy(a), 'axis': axis}

    def _pump(self, target, axis, a):
      return {'a': ut.maybe_copy(a), 'axis': axis}

  # Set the name of the tank class
  TankClass.__name__ = class_name

  # Return the tank_def function.
  def func(a, axis=(), type_dict=None, waterwork=None, name=None, return_tank=False, slot_plugs=None, tube_plugs=None, slot_names=None, tube_names=None):
    tank = TankClass(a=a, axis=axis, waterwork=waterwork, name=name)
    if slot_plugs is not None:
      for key in slot_plugs:
        tank.get_slots()[key].set_plug(slot_plugs[key])
    if tube_plugs is not None:
      for key in tube_plugs:
        tank.get_tubes()[key].set_plug(tube_plugs[key])
    return tank.get_tubes(), tank.get_slots()
  return func



"""Concatenate tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np


class Concatenate(ta.Tank):
  """The defintion of the Concatenate tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'concatenate'
  slot_keys = ['a_list', 'axis']
  tube_keys = ['target', 'axis', 'indices', 'dtypes']
  pass_through_keys = ['axis']

  def _pour(self, a_list, axis):
    """Execute the Concatenate tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a_list: list of arrays
      The list of arrays to concatenate.
    axis: int
      The axis along which to concatenate.

    Returns
    -------
    dict(
      target: np.ndarray
        The concatenation of 'a_list' along 'axis'
      axis: int
        The axis along which to concatenate.
      indices: np.ndarray
        The indices that mark the separation of arrays.
      dtypes: list of dtypes
        The dtypes of the original elements of 'a_list'. Must be of the same length as the orignal 'a_list.'
    )

    """
    indices = []
    dtypes = []
    # Keep a cursor going which represents the current index in the outputted
    # so that it can be split in the right place during pump.
    cursor = 0
    for a_num, a in enumerate(a_list):
      a = np.array(a)

      # Save the dtypes since this can be lots when the concatenation is of
      # separate types.
      dtypes.append(a.dtype)

      # If this is the last item the continue since otherwise it'll create one
      # split too many
      if a_num == len(a_list) - 1:
        continue

      # Append the current value of the cursor plus the size of the dimension.
      indices.append(cursor + a.shape[axis])
      cursor += a.shape[axis]
    return {'target': np.concatenate(a_list, axis=axis), 'indices': np.array(indices), 'axis': axis, 'dtypes': dtypes}

  def _pump(self, target, indices, axis, dtypes):
    """Execute the Concatenate tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The concatenation of 'a_list' along 'axis'
    axis: int
      The axis along which to concatenate.
    indices: np.ndarray
      The indices that mark the separation of arrays.
    dtypes: list of dtypes
      The dtypes of the original elements of 'a_list'. Must be of the same length as the orignal 'a_list.'

    Returns
    -------
    dict(
      a_list: list of arrays
        The list of arrays to concatenate.
      axis: int
        The axis along which to concatenate.
    )

    """
    # Split up the array according the the indices saved from the pour.
    splits = np.split(target, indices, axis=axis)
    a_list = []
    for a, dtype in zip(splits, dtypes):

      # caste them all back to their original types.
      a = a.astype(dtype)
      a_list.append(a)
    return {'a_list': a_list, 'axis': axis}



"""EffectiveLength tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np
import wtrwrks.tanks.utils as ut

class EffectiveLength(ta.Tank):
  """Get the length of the last dimension, not including the default_val.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'effective_length'
  slot_keys = ['a', 'default_val']
  tube_keys = ['target', 'a', 'default_val']
  pass_through_keys = ['a', 'default_val']

  def _pour(self, a, default_val):
    """

    Parameters
    ----------
    a: np.ndarray
      The array to get the effective length of.
    default_val:
      The value to not count

    Returns
    -------
    dict(
      target: np.ndarray
        An array of the same shape as 'a' except missing the last dimension. The values are effective lengths of the last dimesion of a.
      a: np.ndarray
        The array to get the effective length of.
      default_val:
        The value to not count
    )

    """
    zero = (np.array(a) == default_val)

    all_zero = np.all(zero, axis=-1)
    not_zero = ~zero

    reversed_last_dim = not_zero[..., ::-1]

    lengths = np.argmax(reversed_last_dim, axis=-1)
    lengths = a.shape[-1] - lengths
    lengths[all_zero] = 0

    return {'target': lengths, 'a': ut.maybe_copy(a), 'default_val': default_val}

  def _pump(self, target, a, default_val):
    """Execute the Shape tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      An array of the same shape as 'a' except missing the last dimension. The values are effective lengths of the last dimesion of a.
    a: np.ndarray
      The array to get the effective length of.
    default_val:
      The value to not count

    Returns
    -------
    dict(
      a: np.ndarray
        The array to get the effective length of.
      default_val:
        The value to not count
    )

    """

    return {'a': ut.maybe_copy(a), 'default_val': default_val}



"""Reshape tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Remove(ta.Tank):
  """Remove elements of an array according to a mask.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'remove'
  slot_keys = ['a', 'mask']
  tube_keys = ['target', 'mask', 'removed']
  pass_through_keys = ['mask']

  def _pour(self, a, mask):
    """Execute the Flatten tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to execute the remove on
    mask: np.ndarray
      An array of Trues and Falses, telling which elements to remove. Either needs to be the same shape as a or broadcastable to a.

    Returns
    -------
    dict(
      target: np.ndarray
        The array with elements removed
      removed: np.ndarray
        The remove elements of the array
      mask: np.ndarray
        An array of Trues and Falses, telling which elements to remove. Either needs to be the same shape as a or broadcastable to a.
    )

    """
    print a.shape, mask.shape
    target = a[mask]
    removed = a[~mask]
    return {'target': target, 'mask': mask, 'removed': removed}

  def _pump(self, target, mask, removed):
    """Execute the Flatten tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The reshaped array
    old_shape: list of ints
      The old shape of the array

    Returns
    -------
    dict(
      target: np.ndarray
        The array with elements removed
      removed: np.ndarray
        The remove elements of the array
      mask: np.ndarray
        An array of Trues and Falses, telling which elements to remove. Either needs to be the same shape as a or broadcastable to a.
    )

    """
    if target.dtype.type in (np.unicode_, np.string_):
      if target.dtype.itemsize > removed.dtype.itemsize:
        dtype = target.dtype.itemsize
      else:
        dtype = removed.dtype.itemsize
    else:
      dtype = target.dtype

    shape = list(mask.shape)
    shape = shape + list(target.shape[len(shape):])
    a = np.empty(shape, dtype=dtype)

    a[mask] = target
    a[~mask] = removed
    return {'a': a, 'mask': mask}



"""DatasetTransform definition."""
import transform as n
import wtrwrks.tanks.tank_defs as td
import wtrwrks.waterworks.name_space as ns
import wtrwrks.waterworks.waterwork as wa
import wtrwrks.transforms.transform as tr
from wtrwrks.waterworks.empty import empty
import os
import numpy as np
import pandas as pd
import itertools
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
import pprint

class DatasetTransform(tr.Transform):
  """A Transform that is built out of other transform. Used to handle entire dataset which may have varied input types. Transforms an array normalized vectors and any information needed to fully reconstruct the original array.

  Parameters
  ----------
  name : str
    The name of the transform.
  dtype : numpy dtype
    The data type the transformed data should have. Defaults to np.float64.
  input_dtype: numpy dtype
    The datatype of the original inputted array.

  Attributes
  ----------
  input_dtype: numpy dtype
    The datatype of the original inputted array.
  input_shape: list of ints
    The shape of the original inputted array.
  transforms : dict
    The dictionary which holds all the Transforms that the DatasetTransform is built from.
  transform_names : list of strs
    The names of all the subtransforms
  transform_cols : dict
    A mapping from the transform name to a list of strs which define the columns of the datafram that the subtransform acts on.
  params : dict
    A dicitonary which holds any additional information about the dataset the user wants to save. (e.g. num val examples)

  """

  attribute_dict = {'transforms': None, 'transform_cols': None, 'transform_names': None, 'params': None}

  for k, v in n.Transform.attribute_dict.iteritems():
    if k in attribute_dict:
      continue
    attribute_dict[k] = v

  required_params = set()
  required_params.update(n.Transform.required_params)

  def __init__(self, from_file=None, save_dict=None, **kwargs):
    """Define a transform using a dictionary, file, or by setting the attribute values in kwargs.

    Parameters
    ----------
    from_file : None or str
      The file path of the Tranform that was written to disk.
    save_dict : dict or None
      The dictionary of attributes that completely define a Transform.
    name : str
      The name of the transform.
    dtype : numpy dtype
      The data type the transformed data should have. Defaults to np.float64.
    input_dtype: numpy dtype
      The datatype of the original inputted array.
    **kwargs :
      The keyword arguments that set the values of the attributes defined in the attribute_dict.

    """
    super(DatasetTransform, self).__init__(from_file, save_dict, **kwargs)

    if self.transforms is None:
      self.transforms = {}
      self.transform_cols = {}
      self.transform_names = []

    if self.params is None:
      self.params = {}

  def __getitem__(self, key):
    """Return the transform corresponding to key."""
    return self.transforms[key]

  def __iter__(self):
    """Iterate transform set. It is just the iterator of the transforms dictionary."""
    return iter(self.transforms)

  def _calc_global_values(self, array):
    """Calculate all the values of the Transform that are dependent on all the examples of the dataset. (e.g. mean, standard deviation, unique category values, etc.) This method must be run before any actual transformation can be done.

    Parameters
    ----------
    array : np.ndarray
      Some of the data that will be transformed.

    """
    df = pd.DataFrame(
      data=array,
      index=range(array.shape[0]),
      columns=self.cols
    )
    for key in self.transform_names:
      trans = self.transforms[key]

      cols = self.transform_cols[key]
      if np.array(cols).dtype.type not in (np.dtype('O'), np.dtype('S'), np.dtype('U')):
        cols = [self.name + '_' + str(dim) for dim in cols]

      self.transform_cols[key] = cols
      self.transforms[key].cols = cols
      # Get the subarrays and cast them to valid dtypes.
      subarray = df[cols].values
      if trans.input_dtype is None:
        raise ValueError("Must explicitly set the input dtype if using the transform as part of a dataset_transform.")

      subarray = subarray.astype(trans.input_dtype)

      # Calculate the global values for this transform.
      self.transforms[key]._calc_global_values(subarray)
      self.transforms[key].is_calc_run = True

  def _finish_calc(self):
    """Finish up the calc global value process."""
    # Run the finsh calc of it's constituents
    for key in self:
      trans = self.transforms[key]
      trans._finish_calc()

    # Verify that all the columns were used from the array, otherwise throw
    # an error.
    all_ranges = set(range(len(self.cols)))
    for key in self:
      cols = self.transform_cols[key]
      if np.array(cols).dtype.type in (np.dtype('O'), np.dtype('S'), np.dtype('U')):
        cols = [list(self.cols).index(c) for c in cols]
      for index in cols:
        if index in all_ranges:
          all_ranges.remove(index)

    if all_ranges:
      raise ValueError("Must use all columns in array. Columns " + str(sorted([self.cols[i] for i in all_ranges])) + " are unused. Either remove them from the array or all additional transforms which use them.")

  def _from_save_dict(self, save_dict):
    """Reconstruct the transform object from the dictionary of attributes."""
    import wtrwrks.transforms.transform_defs as trd
    for key in self.attribute_dict:
      if key == 'transforms':
        continue
      setattr(self, key, save_dict[key])

    transforms = {}
    for key in save_dict['transforms']:
      trans_save_dict = save_dict['transforms'][key]
      trans = eval('trd.' + trans_save_dict['__class__'])(save_dict=trans_save_dict)
      transforms[key] = trans
    setattr(self, 'transforms', transforms)

  def _get_array_attributes(self, prefix=''):
    """Get the dictionary that contain the original shapes of the arrays before being converted into tfrecord examples.

    Parameters
    ----------
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

    Returns
    -------
    array_attributes : dict
      The dictionary with keys equal to those that are found in the Transform's example dicts and values are the shapes of the arrays of a single example.

    """
    att_dict = {}
    for key in self.transform_names:
      trans = self.transforms[key]
      trans_att_dict = trans._get_array_attributes(prefix=os.path.join(prefix, self.name))
      att_dict.update(trans_att_dict)
    return att_dict

  def _save_dict(self):
    """Create the dictionary of values needed in order to reconstruct the transform."""
    save_dict = {}
    for key in self.attribute_dict:
      if key == 'transforms':
        continue
      save_dict[key] = getattr(self, key)

    save_dict['__class__'] = str(self.__class__.__name__)
    save_dict['__module__'] = str(self.__class__.__module__)

    save_dict['transforms'] = {}
    for key in self.transform_names:
      save_dict['transforms'][key] = self.transforms[key]._save_dict()
    return save_dict

  def _start_calc(self):
    """Start the calc global value process."""
    for key in self:
      trans = self.transforms[key]
      trans._start_calc()
    self.num_examples = 0.

  def add_transform(self, cols, transform):
    """Add another transform to the dataset transform. The added transform must have a unique name.

    Parameters
    ----------
    col_ranges : 2-tuple of ints
      The slice of the array that this transformation will operate on.
    transform : Transform
      The transform object that will operate on the subarray described by col_ranges.

    """
    name = transform.name
    if name is None or name == '':
      raise ValueError("Transform must have it's name set, got: " + str(name))
    elif name in self.transforms:
      raise ValueError(str(name) + " already the name of a transform.")

    self.transform_names.append(name)
    self.transforms[name] = transform
    self.transform_cols[name] = cols
    if len(cols) == 0 or type(cols) is not list:
      raise ValueError("Must pass a non empty list of columns to cols.")
    elif np.array(cols).dtype.type in (np.dtype('O'), np.dtype('S'), np.dtype('U')):
      transform.cols = cols[:]
    else:
      transform.cols = [transform.name + '_' + str(col) for col in cols]

  def define_waterwork(self, array=empty, return_tubes=None, prefix=''):
    """Get the waterwork that completely describes the pour and pump transformations.

    Parameters
    ----------
    array : np.ndarray or empty
      The array to be transformed.
    return_tubes : list of str or None
      Tube objects to be returned from the Waterwork object. Only needed if Waterworks are being stiched together.
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

    Returns
    -------
    Waterwork
      The waterwork with all the tanks (operations) added, and names set.

    """
    assert self.is_calc_run, ("Run calc_global_values before running the transform")

    with ns.NameSpace(self.name):
      indices = []
      for name in self.transform_names:
        trans_cols = self.transform_cols[name]
        indices.append([list(self.cols).index(c) for c in trans_cols])

      # Can only partition along the 0th axis so transpose it so that the
      # 'column' dimension is the first
      perm = [1, 0]
      transp, transp_slots = td.transpose(a=array, axes=perm)

      # Parition the full dataset array into subarrays so that the individual
      # transforms can handle them.
      parts, _ = td.partition_by_index(
        a=transp['target'], indices=indices,
        tube_plugs={
          'missing_cols': np.zeros((0, 1), dtype=self.input_dtype),
          'missing_array': np.zeros((0, 1), dtype=self.input_dtype)
        }
      )
      parts['missing_cols'].set_name('missing_cols')
      parts['missing_array'].set_name('missing_array')
      transp_slots['a'].set_name('array')

      # Split up the Tube object into a list of Tubes so they can each be fed
      # into individual transforms.
      parts_list, _ = td.iter_list(parts['target'], num_entries=len(self.transforms))
      for part, name in zip(parts_list, self.transform_names):
        trans = self.transforms[name]

        # Transpose it back to it's original orientation
        trans_back, _ = td.transpose(a=part, axes=perm, name=name + '-Trans')
        part = trans_back['target']

        # Depending on the type of transform, cast the subarray to its valid
        # type.
        cast, _ = td.cast(
          part, trans.input_dtype,
          tube_plugs={
            'input_dtype': self.input_dtype,
            'diff': np.array([], dtype=self.input_dtype)
          },
          name=name + '-Cast'
        )
        # if isinstance(trans, nt.NumTransform):
        #   cast, _ = td.cast(part, np.float64, name='-'.join([name, 'cast']))
        #   part = cast['target']
        # elif isinstance(trans, dt.DateTimeTransform):
        #   cast, _ = td.cast(part, np.datetime64, name='-'.join([name, 'cast']))
        #   part = cast['target']
        # elif isinstance(trans, st.StringTransform):
        #   cast, _ = td.cast(part, np.unicode, name='-'.join([name, 'cast']))
        #   part = cast['target']
        # elif isinstance(trans, mlst.MultiLingualStringTransform):
        #   cast, _ = td.cast(part, np.unicode, name='-'.join([name, 'cast']))
        #   part = cast['target']
        # elif isinstance(trans, ct.CatTransform):
        #   cast, _ = td.cast(part, np.unicode, name='-'.join([name, 'cast']))
        #   part = cast['target']
        with ns.NameSpace(name):
          trans.define_waterwork(array=cast['target'], prefix=os.path.join(prefix, self.name))

    if return_tubes is not None:
      ww = parts['missing_array'].waterwork
      r_tubes = []
      for r_tube_key in return_tubes:
        r_tubes.append(ww.maybe_get_tube(r_tube_key))
      return r_tubes

  def pour(self, data=None, data_iter=None):
    """Execute the transformation in the pour (forward) direction.

    Parameters
    ----------
    data : np.ndarray
      The numpy array to transform.
    data_iter : iterator of np.array or pd.DataFrame
      The entire dataset in the form of an iterator of numpy array or a pandas DataFrame. Needed if the dataset is too large to fit in memory. Should have the same columns as the arrays that will be fed to the pour method. Can only use if 'data' is not being used
    Returns
    -------
    tap_dict : dict (or iterator of dicts)
      The dictionary of transformed outputs as well as any additional information needed to completely reconstruct the original data. Returns an iterator of dicts if something is passed to 'data_iter' rather than the 'data' argument.

    """

    all_cols = []
    for name in self.transform_names:
      all_cols.extend(self.transform_cols[name])

    # if dealing with an array use numerical columns
    if not type(data) is pd.DataFrame:
      new_all_cols = []
      for col in all_cols:
        new_all_cols.append(self.cols.index(col))
      all_cols = new_all_cols

    def normalize(data):
      if type(data) is pd.DataFrame:
        data = data[all_cols]
      else:
        data = data[:, all_cols]
      return data

    if data is not None and data_iter is None:
      data = normalize(data)
      return super(DatasetTransform, self).pour(data=data)
    elif data_iter is not None and data is None:
      data_iter = itertools.imap(normalize, data_iter)
      return super(DatasetTransform, self).pour(data_iter=data_iter)
    else:
      raise ValueError("Must supply exactly one data or data_iter.")

  def pump(self, tap_dict, df=False, index=None):
    """Execute the transformation in the pump (backward) direction.

    Parameters
    ----------
    kwargs: dict
      The dictionary all information needed to completely reconstruct the original rate.

    Returns
    -------
    array : np.ndarray
      The original numpy array that was poured.

    """
    ww = self.get_waterwork()
    funnel_dict = ww.pump(tap_dict, key_type='str')
    array = funnel_dict[self._pre('array')].astype(self.input_dtype)
    if df:
      if index is None:
        index = np.arange(array.shape[0])
      data = pd.DataFrame(
        data=array,
        index=index,
        columns=self.cols
      )
      for name in self.transform_names:
        trans = self.transforms[name]
        data[trans.cols] = data[trans.cols].astype(trans.input_dtype)
    else:
      data = array

    return data

  def get_eval_class(self, table_name):
    """Create the sqlalchemy class that defines one row of data. Useful for writing evaluation data to a sql database.

    Parameters
    ----------
    table_name : str
      The name of the table that the class will be defined for.

    Returns
    -------
    EvalClass : class
      The class that defines a row of data.

    """
    Base = declarative_base()

    class EvalClass(Base):
      __tablename__ = table_name
      example_id = sa.Column(sa.Integer, primary_key=True)

    already_added_cols = set(['example_id'])
    for name in self.transform_names:
      trans = self.transforms[name]
      for col in trans.cols:

        if col in already_added_cols:
          continue
        already_added_cols.add(col)

        if trans.input_dtype in (np.int64, np.int32, int):
          db_dtype = sa.Integer
        elif trans.input_dtype in (np.float64, np.float32, float):
          db_dtype = sa.Float
        elif trans.input_dtype in (np.bool, bool):
          db_dtype = sa.Boolean
        elif trans.input_dtype in (np.dtype('S'), np.dtype('U'), np.dtype('O')):
          db_dtype = sa.String
        elif trans.input_dtype in (np.datetime64,):
          db_dtype = sa.DateTime
        else:
          raise ValueError("{} is not a supported type to be used on the database.".format())

        setattr(EvalClass, col, sa.Column(db_dtype))

    return EvalClass

  def get_schema_dict(self, var_lim=None):
    """Create a dictionary which defines the proper fields which are needed to store the untransformed data in a (postgres) SQL database.

    Parameters
    ----------
    var_lim : int or dict
      The maximum size of strings. If an int is passed then all VARCHAR fields have the same limit. If a dict is passed then each field gets it's own limit. Defaults to 255 for all string fields.

    Returns
    -------
    schema_dict : dict
      Dictionary where the keys are the field names and the values are the SQL data types.

    """
    schema_dict = {}
    for name in self.transform_names:
      trans = self.transforms[name]

      if type(var_lim) is dict and name in var_lim:
        schema_dict.update(trans.get_schema_dict(var_lim[name]))
      elif type(var_lim) is dict:
        schema_dict.update(trans.get_schema_dict(None))
      else:
        schema_dict.update(trans.get_schema_dict(var_lim))

    return schema_dict

  def get_waterwork(self, array=empty, recreate=False):
    """Create the Transform's waterwork or return the one that was already created.

    Parameters
    ----------
    recreate : bool
      Whether or not to force the transform to create a new waterwork.

    Returns
    -------
    Waterwork
      The waterwork object that this transform creates.

    """
    if self.waterwork is not None and not recreate:
      return self.waterwork
    with wa.Waterwork() as ww:
      self.define_waterwork(array)
    self.waterwork = ww
    return ww



from setuptools import setup, find_packages
from setuptools.command.install import install
import subprocess
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md')) as f:
  long_description = f.read()

print long_description

class CustomDependencyInstallCommand(install):
  """Customized setuptools install command - install all dependencies."""

  @staticmethod
  def _define_dependencies():
    """
    Returns a list of dependency repo urls for pip to download
    Any version updates or additional dependencies should be added here
    """
    dependency_links = [
        'nltk==3.3'
    ]

    return dependency_links

  def run(self):
    """
    Override original run function of install command,
    with actions to install dependencies
    """
    # go through all dependencies defined by user above
    for dep in self._define_dependencies():
      # install them with pip
      subprocess.call('pip install ' + dep, shell=True)

    install.run(self)

    # MIGHT NOT RAISE EXCEPTION IF IT FAILS
    # Install some nltk stuff
    import nltk
    nltk.download('stopwords')
    nltk.download('punkt')
    nltk.download('averaged_perceptron_tagger')
    nltk.download('wordnet')


setup(
  name='wtrwrks',
  description='A system for creating reversible data transformations',
  long_description=long_description,
  long_description_content_type='text/markdown',
  version='2.0.3',
  author='Christopher Silkworth',
  author_email='crsilkworth@gmail.com',
  packages=find_packages(exclude=['unit_test', 'unitTest', '*.unitTest',
                                  'unitTest.*', '*.unitTest.*']),
  install_requires=[
    'numpy~=1.14',
    'pandas~=0.23',
    'matplotlib~=2.2',
    'scipy~=1.1',
    'mpu~=0.14',
    'tables~=3.4',
    'tensorflow~=1.14',
    'nltk~=3.4',
    'tinysegmenter~=0.4',
    'chop~=1.0',
    'konlpy~=0.5',
    'simplejson~=3.16',
    'gensim~=3.6',
    'dill~=0.2',
    'sqlalchemy~=1.3',
    'pathos~=0.2'
  ],
  cmdclass={
      'install': CustomDependencyInstallCommand
  }
)



"""Cast tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np


class Cast(ta.Tank):
  """The defintion of the Cast tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'cast'
  slot_keys = ['a', 'dtype']
  tube_keys = ['target', 'input_dtype', 'diff']

  def _pour(self, a, dtype):
    """Execute the Cast tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The object to be casted to a new type. Must be able to be converted into a numpy array.
    dtype: a numpy dtype
      The type to cast the object to.

    Returns
    -------
    dict(
      target: The type specified by the dtype slot
        The result of casting 'a' to the new dtype.
      input_dtype: a numpy dtype
        The dtype of the original array.
      diff: The datatype of the original 'a' array
        The difference between the original 'a' and the casted array.
    )

    """
    a = np.array(a)
    target = a.astype(dtype)

    # If the inputted is of float type and is being cast to an integer, it will
    # lose information. So the difference between the original and casted array
    # must be saved.
    if a.dtype in (np.float64, np.float32) and dtype in (np.int32, np.int64, np.bool, int):
      diff = a - target
    else:
      diff = np.zeros_like(target)
    return {'target': target, 'diff': diff, 'input_dtype': a.dtype}

  def _pump(self, target, input_dtype, diff):
    """Execute the Cast tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: The type specified by the dtype slot
      The result of casting 'a' to the new dtype.
    input_dtype: a numpy dtype
      The dtype of the original array.
    diff: The datatype of the original 'a' array
      The difference between the original 'a' and the casted array.

    Returns
    -------
    dict(
      a: np.ndarray
        The object to be casted to a new type. Must be able to be converted into a numpy array.
      dtype: a numpy dtype
        The type to cast the object to.
    )

    """
    dtype = target.dtype

    a = target
    if np.issubdtype(input_dtype, np.number):
      a = diff + target
    a = a.astype(input_dtype)
    return {'a': a, 'dtype': dtype}



"""Reshape tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Reshape(ta.Tank):
  """Reshape an array.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'reshape'
  slot_keys = ['a', 'shape']
  tube_keys = ['target', 'old_shape']

  def _pour(self, a, shape):
    """Execute the Flatten tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to reshape
    shape : list of ints
      The new shape of the array.

    Returns
    -------
    dict(
      target: np.ndarray
        The reshaped array
      old_shape: list of ints
        The old shape of the array
    )

    """
    old_shape = a.shape
    target = a.reshape(shape)
    return {'target': target, 'old_shape': old_shape}

  def _pump(self, target, old_shape):
    """Execute the Flatten tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The reshaped array
    old_shape: list of ints
      The old shape of the array

    Returns
    -------
    dict(
      a: np.ndarray
        The array to reshape
      shape : list of ints
        The new shape of the array.
    )

    """
    shape = target.shape
    return {'a': target.reshape(old_shape), 'shape': shape}



import cat_transform
import num_transform
import datetime_transform
import string_transform
import multi_lingual_string_transform
import dataset_transform

CatTransform = cat_transform.CatTransform
NumTransform = num_transform.NumTransform
DateTimeTransform = datetime_transform.DateTimeTransform
StringTransform = string_transform.StringTransform
MultiLingualStringTransform = multi_lingual_string_transform.MultiLingualStringTransform
DatasetTransform = dataset_transform.DatasetTransform



import shutil
import tempfile
import unittest
import wtrwrks.utils.dir_functions as d
import wtrwrks.utils.test_helpers as th
import os
import numpy as np


class TestDirFunctions(unittest.TestCase):
  def setUp(self):
    self.temp_dir = tempfile.mkdtemp()

  def tearDown(self):
    shutil.rmtree(self.temp_dir)

  def test_create_dirs(self):
    dir = os.path.join(self.temp_dir, 'test', 'test')
    d.create_dirs(
        dir=dir
    )

  def test_maybe_create_dir(self):
    out_dir = d.maybe_create_dir(
        self.temp_dir,
        'test',
        'test'
    )

  def test_read_from_file(self):
    temp_json = os.path.join(self.temp_dir, 'temp.json')
    d.save_to_file([], temp_json)

    outputs = d.read_from_file(
        file_name=temp_json
    )

    temp_npy = os.path.join(self.temp_dir, 'temp.npy')
    d.save_to_file(np.zeros(dtype=np.int64, shape=[1]), temp_npy)

    array = d.read_from_file(
        file_name=temp_npy
    )
    th.assert_arrays_equal(self, array, np.zeros(dtype=np.int64, shape=[1]))

  def test_tar_dir(self):
    tar_dir = os.path.join(self.temp_dir, 'temp')
    tar_dir = d.maybe_create_dir(
        tar_dir
    )

    d.tar_dir(
        dir_to_tar=tar_dir,
        dir_to_save_to=self.temp_dir
    )

    d.untar_dir(
        file_to_untar=tar_dir.rstrip('/') + '.tar.gz',
        dir_to_save_to=self.temp_dir
    )


if __name__ == "__main__":
    unittest.main()



"""Flatten tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np

class Flatten(ta.Tank):
  """The defintion of the Flatten tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'flatten'
  slot_keys = ['a']
  tube_keys = ['target', 'shape']

  def _pour(self, a):
    """Execute the Flatten tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to be flattened

    Returns
    -------
    dict(
      target: dtype of 'a'
        The flattened array.
      shape: list of ints.
        The original shape of 'a'.
    )

    """
    shape = a.shape
    target = a.flatten()
    return {'target': target, 'shape': shape}

  def _pump(self, target, shape):
    """Execute the Flatten tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: dtype of 'a'
      The flattened array.
    shape: list of ints.
      The original shape of 'a'.

    Returns
    -------
    dict(
      a: np.ndarray
        The array to be flattened
    )

    """
    return {'a': target.reshape(shape)}



"""Shape tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np
import wtrwrks.tanks.utils as ut
import random


class BertRandomInsert(ta.Tank):
  """A specialized tank built specifically for the BERT ML model. Randomly inserts a [SEP] token at the end of some sentence in a row, then with some probability overwrites the latter part of the string with a randomly selected sentence. For more information or motivation look up the bert model.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'bert_random_insert'
  slot_keys = ['a', 'ends', 'num_tries', 'random_seed']
  tube_keys = ['target', 'removed', 'num_tries', 'ends', 'random_seed', 'segment_ids', 'is_random_next']
  pass_through_keys = ['ends', 'num_tries', 'random_seed']

  def _pour(self, a, ends, num_tries, random_seed):
    """

    Parameters
    ----------
    a: np.ndarray
      The array that will have the [SEP] and [CLS] tags inserted as well as randomly setting half of the rows to having random sentences after the first [SEP] tag.
    ends: np.ndarray of bools
      An array of the same shape as 'a' which marks the end of a sentence with a True.
    num_tries: int
      The number of times to try and find a random sentences to replace the second part of the 'a' array.
    random_seed: int
      The random seed.

    Returns
    -------
    dict(
      target: np.ndarray
        The array a with the [SEP] and [CLS] tags as well a some randomly overwritten second sentences.
      removed: np.ndarray
        A array with the same size as target that contains all the substrings that were overwritten.
      ends: np.ndarray of bools
        An array of the same shape as 'a' which marks the end of a sentence with a True.
      num_tries: int
        The number of times to try and find a random sentences to replace the second part of the 'a' array.
      segment_ids: np.ndarray
        An array of zeros and ones with the same shape as 'a' which says whether the token is part of the first sentence or the second.
      is_random_next: np.ndarray
        An array of bools which says whether the second sentence was replaced with a random sentence.
      random_seed: int
        The random seed.
    )

    """
    np.random.seed(random_seed)
    sepped_array = []
    max_row_len = a.shape[1] + 3

    for row_num in xrange(a.shape[0]):
      sep_index = np.random.choice(np.where(ends[row_num])[0])
      last_sent_index = np.where(ends[row_num])[0][-1]
      row = a[row_num].tolist()
      row.insert(last_sent_index + 1, '[SEP]')
      row.insert(sep_index + 1, '[SEP]')
      row.insert(0, '[CLS]')
      sepped_array.append(row)

    target = []
    removed = []
    segment_ids = []
    is_random_next = []
    for row_num in xrange(len(sepped_array)):
      row = sepped_array[row_num]

      sep_index = row.index('[SEP]')
      segment_ids.append([0] * (sep_index + 1) + [1] * (max_row_len - sep_index - 1))

      removed_chunk = ['[NA]'] * max_row_len
      if np.random.rand() < 0.5:
        space_left = max_row_len - (sep_index + 1)
        chosen_chunk = None
        for _ in xrange(num_tries):
          rand_row = sepped_array[np.random.choice(a.shape[0])]
          chunk = rand_row[rand_row.index('[SEP]') + 1:]
          chunk = chunk[:chunk.index('[SEP]') + 1]
          if len(chunk) < space_left:
            if chosen_chunk is None or len(chunk) > len(chosen_chunk):
              chosen_chunk = chunk
        if chosen_chunk is not None:
          removed_chunk = row[sep_index + 1:]
          removed_chunk = ['[NA]'] * (max_row_len - len(removed_chunk)) + removed_chunk
          row = row[:sep_index + 1] + chosen_chunk
          row = row + [''] * (max_row_len - len(row))

          is_random_next.append(True)
        else:
          is_random_next.append(False)
      else:
        is_random_next.append(False)
      target.append(row)
      removed.append(removed_chunk)
    segment_ids = np.array(segment_ids)
    is_random_next = np.array(is_random_next)
    target = np.array(target)
    removed = np.array(removed)

    return {'target': target, 'removed': removed, 'num_tries': num_tries, 'ends': ends, 'random_seed': random_seed, 'segment_ids': segment_ids, 'is_random_next': is_random_next}

  def _pump(self, target, removed, num_tries, ends, random_seed, segment_ids, is_random_next):
    """Execute the Shape tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The array a with the [SEP] and [CLS] tags as well a some randomly overwritten second sentences.
    removed: np.ndarray
      A array with the same size as target that contains all the substrings that were overwritten.
    ends: np.ndarray of bools
      An array of the same shape as 'a' which marks the end of a sentence with a True.
    num_tries: int
      The number of times to try and find a random sentences to replace the second part of the 'a' array.
    segment_ids: np.ndarray
      An array of zeros and ones with the same shape as 'a' which says whether the token is part of the first sentence or the second.
    is_random_next: np.ndarray
      An array of bools which says whether the second sentence was replaced with a random sentence.
    random_seed: int
      The random seed.

    Returns
    -------
    dict(
      a: np.ndarray
        The array that will have the [SEP] and [CLS] tags inserted as well as randomly setting half of the rows to having random sentences after the first [SEP] tag.
      ends: np.ndarray of bools
        An array of the same shape as 'a' which marks the end of a sentence with a True.
      num_tries: int
        The number of times to try and find a random sentences to replace the second part of the 'a' array.
      random_seed: int
        The random seed.
    )

    """

    mask = removed != '[NA]'
    a = ut.maybe_copy(target)
    a[mask] = removed[mask]

    a = a[~np.isin(a, ['[CLS]', '[SEP]'])]
    a = np.reshape(a, list(target.shape[:-1]) + [target.shape[-1] - 3])

    return {'a': a, 'num_tries': num_tries, 'ends': ends, 'random_seed': random_seed}



import numpy as np
import tensorflow as tf
import codecs

def _int_feat(value):
    """Wrapper for inserting int64 features into Example proto."""
    if isinstance(value, np.ndarray):
        value = value.tolist()
    if not isinstance(value, list) and not isinstance(value, np.ndarray):
        value = [value]
    return tf.train.Feature(int64_list=tf.train.Int64List(value=value))


def _float_feat(value):
    """Wrapper for inserting float64 features into Example proto."""
    if isinstance(value, np.ndarray):
        value = value.tolist()
    if not isinstance(value, list):
        value = [value]
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))


def _bytes_feat(value):
  """Wrapper for inserting bytes features into Example proto."""
  if isinstance(value, np.ndarray):
      dtype = value.dtype
      if dtype.char == 'U':
        value = np.char.encode(value, encoding='utf-8').tolist()
  elif not isinstance(value, list):
    if type(value) in (unicode, np.unicode_):
      value = np.char.encode(value, encoding='utf-8')
    value = [value]
  else:
    if dtype.char == 'U':
      value = np.char.encode(value, encoding='utf-8').tolist()

  return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


def select_feature_func(dtype):
  """Choose a feature function based on the inputted datatype."""
  if dtype in (np.int32, np.int64, np.bool):
    return _int_feat
  elif dtype in (np.float32, np.float64):
    return _float_feat
  elif dtype.type in (np.string_, np.unicode_):
    return _bytes_feat
  else:
    raise TypeError("Only string and number types are supported. Got " + str(dtype))


def select_tf_dtype(dtype):
  """Choose a tensorflow type based on the inputted numpy datatype."""
  if dtype in (np.int32, np.int64, np.bool):
    return tf.int64
  elif dtype in (np.float32, np.float64):
    return tf.float32
  elif dtype.type in (np.string_, np.unicode_):
    return tf.string
  else:
    raise TypeError("Only string and number types are supported. Got " + str(dtype))


def size_from_shape(shape):
  """Calculate the size of the array from it's shape."""
  if not shape:
    return 1

  return np.prod(shape)



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestCatToIndex(th.TestTank):
  def test_int(self):
    cat_to_index_map = {1: 0, 2: 1, 3: 2}
    self.pour_pump(td.cat_to_index, {'cats': 1, 'cat_to_index_map': cat_to_index_map}, {'target': 0, 'missing_vals': 0, 'cat_to_index_map': cat_to_index_map, 'input_dtype': int}, type_dict={'cats': int, 'cat_to_index_map': dict})

  def test_float(self):
    cat_to_index_map = {1.: 0, 2.: 1, 3.: 2}
    self.pour_pump(td.cat_to_index, {'cats': 1., 'cat_to_index_map': cat_to_index_map}, {'target': 0, 'missing_vals': 0.0, 'cat_to_index_map': cat_to_index_map, 'input_dtype': float}, type_dict={'cats': int, 'cat_to_index_map': dict})

  def test_str(self):
    cat_to_index_map = {'a': 0, 'b': 1, 'c': 2}
    self.pour_pump(
      td.cat_to_index,
      {'cats': 'b', 'cat_to_index_map': cat_to_index_map},
      {'target': 1, 'missing_vals': '', 'cat_to_index_map': cat_to_index_map, 'input_dtype': str},
      type_dict={'cats': str, 'cat_to_index_map': dict}
    )

    cat_to_index_map = {'a': 0, 'b': 1, 'c': 2}
    self.pour_pump(td.cat_to_index, {'cats': 'd', 'cat_to_index_map': cat_to_index_map}, {'target': -1, 'missing_vals': 'd', 'cat_to_index_map': cat_to_index_map, 'input_dtype': str}, type_dict={'cats': str, 'cat_to_index_map': dict})

  def test_scalar(self):
    cat_to_index_map = {'a': 0, 'b': 1, 'c': 2}
    self.pour_pump(
      td.cat_to_index,
      {'cats': np.array('d'), 'cat_to_index_map': cat_to_index_map},
      {'target': np.array(-1), 'missing_vals': 'd', 'cat_to_index_map': cat_to_index_map, 'input_dtype': np.object},
      type_dict={'cats': np.ndarray, 'cat_to_index_map': dict}
    )

  def test_one_d(self):
    cat_to_index_map = {0.0: 0, 1.5: 1, 3.0: 2}
    self.pour_pump(
      td.cat_to_index,
      {'cats': np.array([2., 3.]), 'cat_to_index_map': cat_to_index_map},
      {'target': np.array([-1, 2]), 'missing_vals': [2.0, 0.0], 'cat_to_index_map': cat_to_index_map, 'input_dtype': np.float64},
      type_dict={'cats': np.ndarray, 'cat_to_index_map': dict}
    )

  def test_two_d(self):
    cat_to_index_map = {'a': 0, 'b': 1, 'c': 2, 'd': 3}
    self.pour_pump(
      td.cat_to_index,
      {'cats': np.array([['a', 'b'], ['c', 'd'], ['e', 'f'], ['a', 'h']]), 'cat_to_index_map': cat_to_index_map},
      {
        'target': np.array([[0, 1], [2, 3], [-1, -1], [0, -1]]),
        'missing_vals': [['', ''], ['', ''], ['e', 'f'], ['', 'h']],
        'cat_to_index_map': cat_to_index_map,
        'input_dtype': np.object
      },
      type_dict={'cats': np.ndarray, 'cat_to_index_map': dict}
    )

if __name__ == "__main__":
    unittest.main()



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestOneHot(th.TestTank):
  def test_int(self):
    self.pour_pump(td.one_hot, {'indices': 1, 'depth': 2}, {'target': np.array([0., 1.]), 'missing_vals': -2}, type_dict={'indices': int, 'depth': int}, test_type=False)

    self.pour_pump(td.one_hot, {'indices': 3, 'depth': 2}, {'target': np.array([0, 0]), 'missing_vals': 3}, type_dict={'indices': int, 'depth': int}, test_type=False)

  def test_scalar(self):
    self.pour_pump(
      td.one_hot,
      {'indices': np.array(3), 'depth': 2},
      {'target': np.array([0., 0.]), 'missing_vals': 3},
      type_dict={'indices': np.ndarray, 'depth': int},
      test_type=False
    )

  def test_one_d(self):
    self.pour_pump(
      td.one_hot,
      {'indices': np.array([2, 3]), 'depth': 3},
      {'target': np.array([[0., 0., 1.], [0, 0, 0]]), 'missing_vals': np.array([-2, 3])},
      type_dict={'indices': np.ndarray, 'depth': int}
    )

  def test_two_d(self):
    self.pour_pump(
      td.one_hot,
      {'indices': np.array([[0, 1], [2, 3], [4, 5], [1, 0]]), 'depth': 3},
      {
        'target': np.array(
          [
            [[1., 0, 0], [0, 1, 0]],
            [[0, 0, 1], [0, 0, 0]],
            [[0, 0, 0], [0, 0, 0]],
            [[0, 1, 0], [1, 0, 0]],
          ]
        ),
        'missing_vals': np.array([[-2, -2], [-2, 3], [4, 5], [-2, -2]])
      },
      type_dict={'indices': np.ndarray, 'depth': int}
    )

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np
import tinysegmenter
from nltk.stem.wordnet import WordNetLemmatizer
from nltk.corpus import wordnet
import nltk

word_net_lemmatizer = WordNetLemmatizer()


def get_wordnet_pos(treebank_tag):
  if treebank_tag.startswith('J'):
    return wordnet.ADJ
  elif treebank_tag.startswith('V'):
    return wordnet.VERB
  elif treebank_tag.startswith('N'):
    return wordnet.NOUN
  elif treebank_tag.startswith('R'):
    return wordnet.ADV
  else:
    return wordnet.NOUN


def basic_lemmatizer(string):
  if string == '':
    return string

  pos_tags = nltk.pos_tag([string])

  if not pos_tags:
    return string

  pos_tag = get_wordnet_pos(pos_tags[0][1])
  lemma = word_net_lemmatizer.lemmatize(string, pos_tag)

  return lemma


class TestLemmatize(th.TestTank):
  def test_scalar(self):
    strings = "running"
    diff = '[["i", 3, 3, "ning"]]'
    self.pour_pump(
      td.lemmatize,
      {
        'strings': strings,
        'lemmatizer': basic_lemmatizer,
      },
      {
        'target': np.array(u'run'),
        'diff': diff,
        'lemmatizer': basic_lemmatizer,
      },
      test_type=False
    )

  def test_one_d(self):
    lemmatizer = lambda s: s
    diff = [['[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]'], ['[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]']]
    strings = np.array([[u'\u30c1\u30e9\u30b7', u'\u30fb', u'\u52e7\u8a98', u'\u5370\u5237', u'\u7269', u'\u306e', u'\u7121\u65ad', u'\u6295\u51fd', u'\u306f', u'\u4e00'], [u'\u3059\u307f\u307e\u305b', u'\u3093', u'\u304c', u'\u3001', u'\u3082\u3046', u'\u4e00', u'\u5ea6', u'\u3069\u304a\u9858\u3044', u'\u3057', u'\u307e\u3059']], dtype=np.unicode)
    target = np.array([[u'\u30c1\u30e9\u30b7', u'\u30fb', u'\u52e7\u8a98', u'\u5370\u5237', u'\u7269', u'\u306e', u'\u7121\u65ad', u'\u6295\u51fd', u'\u306f', u'\u4e00'], [u'\u3059\u307f\u307e\u305b', u'\u3093', u'\u304c', u'\u3001', u'\u3082\u3046', u'\u4e00', u'\u5ea6', u'\u3069\u304a\u9858\u3044', u'\u3057', u'\u307e\u3059']], dtype=np.unicode)
    self.pour_pump(
      td.lemmatize,
      {
        'strings': strings,
        'lemmatizer': lemmatizer,
      },
      {
        'target': target,
        'diff': diff,
        'lemmatizer': lemmatizer,
      },
      test_type=False
    )

  def test_two_d(self):

    diff = [
      [
        ['[]', '[["i", 0, 2, "is"]]', '[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]'],
        ['[]', '[]', '[]', '[["i", 2, 4, "s"]]', '[["i", 3, 3, "n"]]', '[]', '[]', '[]', '[]', '[]']],
      [
        ['[]', '[]', '[["i", 0, 2, "is"]]', '[]', '[]', '[]', '[]', '[]', '[]', '[]'],
        ['[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]', '[]']
      ]
    ]

    strings = np.array(
      [
        [
          ['It', 'is', 'what', 'it', 'is.', '', '', '', '', ''], ['Whatever,', "Bob's", 'mother', 'has', 'seen', 'the', 'world.', '', '', '']
        ],
        [
          ['The', 'sun', 'is', 'not', 'yellow,', "it's", 'chicken.', 'OK.', '', ''],
          ["Don't", 'need', 'a', 'weatherman', 'to', 'know', 'which', 'way', 'the', 'wind']
        ]
      ])
    target = np.array([[['It', 'be', 'what', 'it', 'is.', '', '', '', '', ''], ['Whatever,', "Bob's", 'mother', 'have', 'see', 'the', 'world.', '', '', '']], [['The', 'sun', 'be', 'not', 'yellow,', "it's", 'chicken.', 'OK.', '', ''], ["Don't", 'need', 'a', 'weatherman', 'to', 'know', 'which', 'way', 'the', 'wind']]])
    self.pour_pump(
      td.lemmatize,
      {
        'strings': strings,
        'lemmatizer': basic_lemmatizer,
      },
      {
        'target': target,
        'diff': diff,
        'lemmatizer': basic_lemmatizer,
      },
      test_type=False
    )

if __name__ == "__main__":
    unittest.main()
